/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * <BR><BR>
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * <BR><BR>
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * <BR><BR>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * <BR><BR>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * <BR><BR>
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */

 




package org.greenstone.gatherer.cdm;
/**************************************************************************************
 * Title:        Gatherer
 * Description:  The Gatherer: a tool for gathering and enriching a digital collection.
 * Copyright:    Copyright (c) 2001
 * Company:      The University of Waikato
 * Written:      22/05/02
 * Revised:      17/11/02 - Commented
 **************************************************************************************/
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.Collections;
import javax.swing.BorderFactory;
import javax.swing.DefaultListModel;
import javax.swing.ListSelectionModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableColumn;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.cdm.CollectionDesignManager;
import org.greenstone.gatherer.cdm.CollectionMeta;
import org.greenstone.gatherer.cdm.CollectionMetaManager;
import org.greenstone.gatherer.cdm.Language;
import org.greenstone.gatherer.cdm.LanguageManager;
/** This class provides a graphical interface to allow a user to quickly and conviently (ie all in one place) translate the text fragments associated with general metadata and indexes into each of the assigned languages in the collection. It should provide clear controls for the editing of these text fragments, plus clear indicate what languages still need further translation, which it will do through a combination of coloring and other visual indicators.
 * @author John Thompson, Greenstone Digital Library, University of Waikato
 * @version 2.3
 */
public class TranslationManager
    extends JDialog {
    /** A reference to our creator, the CollectionDesignManager. */
    private CollectionDesignManager manager = null;
    /** The currently selected collection metadata. */
    private CollectionMeta current = null;
    /** A listener for selection changes in the feature list. */
    private FeatureListener features_listener = null;
    /** A reference to the Gatherer, for access to the Dictionary and messaging purposes. */
    private Gatherer gatherer = null;
    /** The button used to close the translation dialog. */
    private JButton close_button = null;
    /** A reference to the dialog so that our inner classes can interact with it. */
    private JDialog dialog = null;
    /** The label denoting the features list. */
    private JLabel features_label = null;
    /** The label denoting the language table. */
    private JLabel languages_label = null;
    /** The label denoting the original text area. */
    private JLabel original_label = null;
    /** The label denoting the original language field */
    private JLabel original_language_label = null;
    /** The label denoting the translated text area. */
    private JLabel translation_label = null;
    /** The label denoting the translated language field. */
    private JLabel translation_language_label = null;
    /** A list of the various features that have collection metadata associated with them. */
    private JList features_list = null;
    /** The base panel. */
    private JPanel content_pane = null;
    /** The right panel. */
    private JPanel control_pane = null;
    /** The panel the features list is shown in. */
    private JPanel features_pane = null;
    /** The panel the languages table is shown in. */
    private JPanel languages_pane = null;
    /** The panel the original text is shown in. */
    private JPanel original_pane = null;
    /** The panel the original language field is shown in. */
    private JPanel original_header_pane = null;
    /** The panel the translation text is shown in. */
    private JPanel translation_pane = null;
    /** The panel the translation language field is shown in. */
    private JPanel translation_header_pane = null;
    /** The table used to show the languages supported by the collection. */
    private JTable languages_table = null;
    /** The text area displaying the original text (in the default language). */
    private JTextArea original = null;
    /** The text area displaying the translated text (in the selected language). */
    private JTextArea translation = null;
    /** The language currently selected. */
    private Language current_language = null;
    /** The proxy model to the language manager which serves as the basis for the table. */
    private TableProxy languages_model = null;
    /** The default size of the features list. */
    static final private Dimension FEATURE_SIZE = new Dimension(200, 400);
    /** The default size of the translation dialog. */
    static final private Dimension SIZE = new Dimension(800, 450);
    /** Constructor.
     * @param gatherer A reference to the <strong>Gatherer</strong>.
     * @param manager A reference to the <strong>CollectionDesignManager</strong>.
     * @see org.greenstone.gatherer.Configuration
     * @see org.greenstone.gatherer.cdm.CollectionMeta
     * @see org.greenstone.gatherer.cdm.TranslationManager.TableProxy
     */
    public TranslationManager(Gatherer gatherer, CollectionDesignManager manager) {
	super();
	this.dialog = this;
	this.gatherer = gatherer;
	this.manager = manager;
	this.languages_model = new TableProxy();
	// Create the features model. Basically we want one string entry for each unique name in the metadata.
	ArrayList features_model = new ArrayList();
	for(int i = 0; i < manager.collectionmetadatum.size(); i++) {
	    CollectionMeta metadata = (CollectionMeta) manager.collectionmetadatum.get(i);
	    Object object = metadata.getName();
	    String name = null;
	    if(object instanceof Index) {
		name = ((Index)object).toString(false);
	    }
	    else {
		name = object.toString();
	    }
	    if(!features_model.contains(name)) {
		features_model.add(name);
	    }
	}
	Collections.sort(features_model);
	// Create components...
	setModal(true);
	setSize(SIZE);
	close_button = new JButton(get("General.Close"));
	content_pane = (JPanel) getContentPane();
	control_pane = new JPanel();
	features_label = new JLabel(get("Affected_Features"));
	features_label.setHorizontalAlignment(JLabel.CENTER);
	features_list = new JList(features_model.toArray());
	features_listener = new FeatureListener();
	features_pane = new JPanel();
	features_pane.setPreferredSize(FEATURE_SIZE);
	languages_table = new JTable(languages_model);
	languages_table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	languages_label = new JLabel(get("Assigned_Languages"));
	languages_label.setHorizontalAlignment(JLabel.CENTER);
	languages_pane = new JPanel();
	original = new JTextArea();
	original.setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
	original.setEditable(false);
	original.setFont(new Font("Arial Unicode MS", Font.PLAIN, 10));
	original_header_pane = new JPanel();
	original_label = new JLabel(get("Default_Text"));
	original_label.setHorizontalAlignment(JLabel.CENTER);
	original_language_label = new JLabel("");
	original_pane = new JPanel();
	translation = new JTextArea();
	translation.setEnabled(false);
	translation.setFont(new Font("Arial Unicode MS", Font.PLAIN, 10));
	translation_header_pane = new JPanel();
	translation_label = new JLabel(get("Translation"));
	translation_label.setHorizontalAlignment(JLabel.CENTER);
	translation_language_label = new JLabel("");
	translation_pane = new JPanel();
	// Add listeners...
	close_button.addActionListener(new CloseListener());
	translation.addKeyListener(new TranslationListener());
	features_list.addListSelectionListener(features_listener);
	features_list.addListSelectionListener(languages_model);
	ListSelectionModel languages_selection_model = languages_table.getSelectionModel();
	languages_selection_model.addListSelectionListener(new TableListener());
	// Layout...
	features_pane.setBorder(BorderFactory.createEmptyBorder(0,0,0,5));
	features_pane.setLayout(new BorderLayout());
	features_pane.add(features_label, BorderLayout.NORTH);
	features_pane.add(new JScrollPane(features_list), BorderLayout.CENTER);
	languages_label.setBorder(BorderFactory.createEmptyBorder(5,0,5,0));
	languages_pane.setLayout(new BorderLayout());
	languages_pane.add(languages_label, BorderLayout.NORTH);
	languages_pane.add(new JScrollPane(languages_table), BorderLayout.CENTER);
	original_header_pane.setBorder(BorderFactory.createEmptyBorder(5,0,5,0));
	original_header_pane.setLayout(new BorderLayout());
	original_header_pane.add(original_label, BorderLayout.WEST);
	original_header_pane.add(original_language_label, BorderLayout.EAST);
	original_pane.setLayout(new BorderLayout());
	original_pane.add(original_header_pane, BorderLayout.NORTH);
	original_pane.add(new JScrollPane(original), BorderLayout.CENTER);
	translation_header_pane.setBorder(BorderFactory.createEmptyBorder(5,0,5,0));
	translation_header_pane.setLayout(new BorderLayout());
	translation_header_pane.add(translation_label, BorderLayout.WEST);
	translation_header_pane.add(translation_language_label, BorderLayout.EAST);
	translation_pane.setLayout(new BorderLayout());
	translation_pane.add(translation_header_pane, BorderLayout.NORTH);
	translation_pane.add(new JScrollPane(translation), BorderLayout.CENTER);
	control_pane.setLayout(new GridLayout(3,1));
	control_pane.add(languages_pane);
	control_pane.add(original_pane);
	control_pane.add(translation_pane);
	content_pane.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
	content_pane.setLayout(new BorderLayout());
	content_pane.add(features_pane, BorderLayout.WEST);
	content_pane.add(control_pane, BorderLayout.CENTER);
	content_pane.add(close_button, BorderLayout.SOUTH);
	// Display...
	Dimension screen_size = Toolkit.getDefaultToolkit().getScreenSize();
	for (int i = 0; i < languages_model.getColumnCount() - 1; i++) {
	    TableColumn column = languages_table.getColumnModel().getColumn(i);
	    column.setMinWidth(0);
	    column.setMaxWidth(screen_size.width);
	    column.setPreferredWidth((SIZE.width - FEATURE_SIZE.width) / 6);
	}
	setLocation((screen_size.width - SIZE.width) / 2, (screen_size.height - SIZE.height) / 2);
	show();
    }
    /** Destructor
	  */
    public void destroy() {
	dialog = null;
    }
    /** Retrieve a phrase from the dictionary based on a certain key.
	  * @param key The search <strong>String</strong>.
	  * @return The matching phrase from the Dictionary as a <strong>String</strong>.
	  */
    private String get(String key) {
	return get(key, null);
    }
    /** Retrieve a phrase from the dictionary based on a certain key and certain arguments.
	  * @param key The search <strong>String</strong>.
	  * @param args A <strong>String[]</strong> used to complete and format the returned phrase.
	  * @return The matching phrase from the Dictionary as a <strong>String</strong>.
	  * @see org.greenstone.gatherer.Dictionary
	  * @see org.greenstone.gatherer.Gatherer
	  */
    private String get(String key, String args[]) {
	if(key.indexOf(".") == -1) {
	    key = "CDM.TranslationManager." + key;
	}
	return gatherer.dictionary.get(key, args);
    }
    /** Listens for actions apon the close button. */
    private class CloseListener
	implements ActionListener {
	/** Closes the dialog. Any implementation of an ActionListener must include this method so we can be informed when an action has occured.
	 * @param event An <Strong>ActionEvent</strong> containing details about the action.
			*/
	public void actionPerformed(ActionEvent event) {
	    dialog.dispose();
	}
    }
    /** Listens to selections within the feature list. */
    private class FeatureListener
	implements ListSelectionListener {
	/** Called whenever the selection changes in the registered list, which requires us to create a new TableProxy.
	 * @param event A <strong>ListSelectionEvent</strong> containing more information about the list selection change.
	 * @see org.greenstone.gatherer.cdm.CollectionMeta
	 * @see org.greenstone.gatherer.cdm.Language
	 */
	public void valueChanged(ListSelectionEvent event) {
				// Try to find the default language value for this feature.
	    String name = (String) features_list.getSelectedValue();
	    Language language = manager.languages.getDefaultLanguage();
	    CollectionMeta metadata = manager.collectionmetadatum.getMetadata(name, language, true);
	    if(metadata != null && metadata.getValue() != null && metadata.getValue().length() != 0) {
		if(!metadata.getLanguage().equals(language)) {
		    original_label.setText(get("Existing_Text"));
		}
		else {
		    original_label.setText(get("Default_Text"));
		}
		original.setText(metadata.getValue());
		original_language_label.setText(metadata.getLanguage().toString());
	    }
	    else {
		original_label.setText(get("No_Text"));
		original.setText("");
	    }
	}
    }
    /** Listens for selections within the languages table. */
    private class TableListener
	implements ListSelectionListener {
	/** Called whenever the selection changes in the registered list, which requires us to update several parts of the dialog.
	 * @param event A <strong>ListSelectionEvent</strong> containing more information about the list selection change.
	 * @see org.greenstone.gatherer.cdm.CollectionMeta
	 * @see org.greenstone.gatherer.cdm.Language
	 */
	public void valueChanged(ListSelectionEvent event) {
	    if (event.getValueIsAdjusting()) return;
	    ListSelectionModel lsm = (ListSelectionModel)event.getSource();
	    if (!lsm.isSelectionEmpty()) {
		int row = lsm.getMinSelectionIndex();
		current = languages_model.getMetadata(row);
		if(current != null) {
		    if (current.getValue() != null) {
			translation.setText(current.getValue());
		    }
		    else {
			translation.setText("");
		    }
		    current_language = current.getLanguage();
		}
		else {
		    current_language = (Language) languages_model.getValueAt(row, 1);
		    translation.setText("");
		}
		features_listener.valueChanged(null);
		translation_language_label.setText(current_language.toString());
		translation.setEnabled(true);
	    }
	}
    }
    /** A AbstractTableModel extension based upon a vector of CollectionMeta, which can also listen to a list for selection changes. */
    private class TableProxy
	extends AbstractTableModel
	implements ListSelectionListener {
	/** Called when the contents of a cell changes.
	 */
	public void cellChanged() {
	    fireTableDataChanged();
	}
	/** Returns the most specific superclass for all the cell values in the column.
			* @param column An <i>int</i> specifying which columns class to be determined.
			* @return The columns <strong>Class</strong>.
			*/
        public Class getColumnClass(int column) {
	    return String.class;
	}
	/** Returns the number of columns in the model.
			* @return An <i>int</i> specifying the number of columns.
			*/
	public int getColumnCount() {
	    if(!features_list.isSelectionEmpty()) {
		return 2;
	    }
	    return 0;
	}
        /** Returns the name of the column at columnIndex.
			* @param column An <i>int</i> specifying the column whose name you are interesting in.
			* @return The name of the column as a <strong>String</strong>.
			*/
	public String getColumnName(int column) {
	    String name = null;
	    switch(column) {
	    case 0:
		name = get("Language");
		break;
	    case 1:
		name = get("Text");
		break;
	    }
	    return name;
	}
	/** Retrieve the collection metadata from the specified row.
		  * @param row An <i>int</i> indicating the row in question.
		  * @see org.greenstone.gatherer.cdm.Language
		  * @see org.greenstone.gatherer.cdm.LanguageManager
		  */
	public CollectionMeta getMetadata(int row) {
	    String name = (String) features_list.getSelectedValue();
	    Language language = (Language) getValueAt(row, 0);
	    return manager.collectionmetadatum.getMetadata(name, language, false);
	}
        /** Returns the number of rows in the model.
			* @return An <i>int</i> specifying the number of rows in the table, which is exactly the number of languages assigned to this collection.
			* @see org.greenstone.gatherer.cdm.LanguageManager
			*/
	public int getRowCount() {
	    return manager.languages.size();
	}
	/** Returns the value for the cell at column and row.
			* @param row The desired row as an <i>int</i>.
			* @param column The desired column as an <i>int</i>.
			* @return The value at the row, column reference.
			* @see org.greenstone.gatherer.cdm.CollectionMeta
			* @see org.greenstone.gatherer.cdm.Language
			* @see org.greenstone.gatherer.cdm.LanguageManager
			*/
	public Object getValueAt(int row, int column) {
	    if(!features_list.isSelectionEmpty()) {
		// First of all find the Language that matches the row and the current feature name, because in any case you'll need them.
		Language language = (Language)manager.languages.get(row);
		String feature_name = (String)features_list.getSelectedValue();
		// Also try to locate the collection level metadata that matches.
		CollectionMeta metadata = manager.collectionmetadatum.getMetadata(feature_name, language, false);
		// Now we do subtly different things depending on the column.
		switch(column) {
		case 0: // The Language. Easiest case.
		    return language;
		case 1: // The value of the metadata. Needs us to retrieve the value for a certain feature and language.
		    if(metadata != null) {
			return metadata.getValue();
		    }
		}
		// If all else fails, return an empty string and hope the class caster doesn't explode.
	    }
	    return "";
	}
        /** Returns true if the cell at row and column is editable.
			* @param row The row of the cell as an <i>int</i>.
			* @param column The column of the cell also as an <i>int</i>.
			*/
	public boolean isCellEditable(int rowIndex, int columnIndex) {
	    return false;
	}
        /** Sets the value in the cell at columnIndex and rowIndex to aValue.
			*/
	public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
	}
	/** Any implementation of <i>ListSelectionListener</i> must include this method so we can be informed when the list selection has changed.
			* @param event A <strong>ListSelectionEvent</strong> containing information about the selection change.
			*/
	public void valueChanged(ListSelectionEvent event) {
	    fireTableStructureChanged();
	}
    }
    /** This provides a custom renderer that colours the background of the default language cell green, and the foreground of currently unready metadata rows red. */
    private class TableRenderer
	extends DefaultTableCellRenderer {
	/** Retrieve the component used to rubberstamp the table based on the given parameters.
	 * @param table The <strong>JTable</strong> to rendered.
	 * @param value The <strong>Object</strong> whose representation will be shown in the table row.
	 * @param isSelected <i>true</i> iff the column/row is selected.
	 * @param hasFocus <i>true</i> iff the column/row is in focus.
	 * @param row An <i>int</i> specifying the target row.
	 * @param column An <i>int</i> specifying the target column.
	 * @see org.greenstone.gatherer.cdm.Language
	 */
	public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
	    Component component = null;
				// Create the component.
	    if(value instanceof String || value instanceof Language) {
		JLabel label = new JLabel(value.toString());
		label.setOpaque(true);
		component = label;
	    }
	    else {
		JCheckBox check = new JCheckBox("", ((Boolean)value).booleanValue());
		JPanel pane = new JPanel(new BorderLayout());
		pane.add(check, BorderLayout.CENTER);
		component = pane;
	    }
				// Check for default property
	    Language language = (Language)languages_model.getValueAt(row, 1);
	    if(language.isDefault()) {
		if(isSelected) {
		    component.setBackground(Gatherer.config.getColor("coloring.collection_selection_background", false));
		}
		else {
		    component.setBackground(Gatherer.config.getColor("coloring.collection_tree_background", false));
		}
	    }
	    else {
		if(isSelected) {
		    component.setBackground(table.getSelectionBackground());
		}
		else {
		    component.setForeground(Color.white);
		}
	    }
				// Check for readiness
	    Boolean ready = (Boolean)languages_model.getValueAt(row, 0);
	    if(!ready.booleanValue()) {
		component.setForeground(Color.red);
	    }
	    else {
		component.setForeground(Color.black);
	    }
	    return component;
	}
    }
    /** Listens for actions within the translation text field and immediately updates the collection metadata as necessary. */
    private class TranslationListener
	extends KeyAdapter {
	/** This method is called when a key has been released, and is perfect for us to synchronize the contents of the translation field with the collection metadata associated with the selected feature and language.
	 * @param event A <strong>KeyEvent</strong> containing more detail about the key released.
	 * @see org.greenstone.gatherer.cdm.CollectionMeta
	 * @see org.greenstone.gatherer.cdm.CollectionMetaManager
	 * @see org.greenstone.gatherer.cdm.Index
	 */
	public void keyReleased(KeyEvent event) {
	    if(current == null) {
		String name = (String)features_list.getSelectedValue();
		Index index = manager.indexes.getIndex(name);
		if(index == null) {
		    current = new CollectionMeta(manager, name, current_language, translation.getText());
		}
		else {
		    current = new CollectionMeta(manager, index, current_language, translation.getText());
		}
		manager.collectionmetadatum.addMetadata(current);
	    }
	    else {
		current.update(current.getName(), current.getLanguage(), translation.getText());
	    }
	    languages_model.cellChanged();
	}
    }
}
