/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * <BR><BR>
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * <BR><BR>
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * <BR><BR>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * <BR><BR>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * <BR><BR>
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */

 




package org.greenstone.gatherer.gui.browser;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.net.URL;
import java.util.Vector;
import javax.swing.JEditorPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.event.EventListenerList;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.util.GURL;
import org.greenstone.gatherer.util.Utility;
/**
 * Class Name:  GHTMLPane
 * Description: To provide a most basic implementation of the HTMLViewPane. Its
 *              rendering performance may be poor, but it will fully support all
 *              methods defined in the interface.
 */
public class GHTMLPane
    extends JPanel
    implements HTMLViewPane, HyperlinkListener, PropertyChangeListener {
	 
    /** follow_links - does the class automatically follow hyperlinks. */
    boolean follow_links = true;
    /** ready - has the view finished rendering the current URL. */
    boolean ready = true;
    /** the list of listeners associated with this class. */
    EventListenerList listeners = null;
    /** gatherer - a reference to the main class of Gatherer for messaging. */
    Gatherer gatherer = null;
    /** view - the JEditorPane the page content will be rendered on to. */
    JEditorPane view = null;
    Object control = null;
    /** current_page - the URL of the currently shown page. */
    GURL current_url = null;
    /** back_pages - Vector of pages that have been visited. */
    Vector back_pages = null;
    /** Vector of pages that have been backed up from. 
     *  Reset whenever a new url 'path' is entered.
     */
    Vector forward_pages = null;


    /** Constructor - collects initial settings, lays out view, renders initial
     *  URL.
     *  @param gatherer - the main class of Gatherer
     */
    public GHTMLPane(Gatherer gatherer, Object control) {
	this.gatherer = gatherer;
	this.control = control;
	this.setLayout(new BorderLayout());

	listeners = new EventListenerList();

	back_pages = new Vector();
	forward_pages = new Vector(); 

	gatherer.debug("\t\tCreating GHTMLPane");
	view = new JEditorPane();
	view.setEditable(false);
	view.addHyperlinkListener(this);
	view.addPropertyChangeListener(this);
	this.add(new JScrollPane(view), BorderLayout.CENTER);
    }

    public void addGBrowserListener(GBrowserListener listener) {
	listeners.add(GBrowserListener.class, listener);
    }

    public void afterDisplay() {
	// Nothing to do. Only realy for embedded apps.
    }

    public boolean back() {
	// Add current page to forward
	forward_pages.add(0,current_url);
	// Back up one page.
	current_url = (GURL) back_pages.lastElement();
	back_pages.remove(back_pages.size() - 1);
	this.setPage(current_url);
	return true;
    }

    public boolean canBack() {
	if(back_pages.size() > 0 && ready) {
	    return true;
	}
	return false;
    }

    public boolean canForward() {
	if(forward_pages.size() > 0 && ready) {
	    return true;
	}
	return false;
    }

    public boolean canStop() {
	if(!ready) {
	    return true;
	}
	return false;
    }

    // This code courtesy of Horst (hheister@epal.com) who provides this solution
    // to the crap null exception I keep getting when I try to grab the image
    // of a rendered HTML.
    public BufferedImage capture() {
	if(ready) {
				// Paint to an off screen buffer
	    Dimension size = view.getSize();
				//Graphics offgraphics = offscreen.getGraphics();
	    BufferedImage offscreen = 
		new BufferedImage(size.width, size.height, 
				  BufferedImage.TYPE_INT_RGB);
	    Graphics2D offgraphics = (Graphics2D) offscreen.getGraphics();
				// For some reason the clip rect is null for the created image
				// (usually it is set to the size of the image) so set it.
	    offgraphics.setClip(0, 0, size.width, size.height);
	    offgraphics.setColor(getBackground());
	    offgraphics.fillRect(0, 0, size.width, size.height);
				// This will display components in the HTML Page
	    view.paint(offgraphics);
	    return offscreen;
	}
	return null;
    }
		  
    public GURL current() {
	return current_url;
    }

    public void followHyperlinks(boolean value) {
	follow_links = value;
    }

    public boolean forward() {
	// Add current page to back
	back_pages.add(current_url);
	// Show forward page.
	current_url = (GURL) forward_pages.firstElement();
	forward_pages.remove(0);
	setPage(current_url);
	return true;
    }

    /** Since you can't add an interface to a component hierarchy this method is required to get a reference of the actual rendering component. 
     * @return This HTMLViewPane implementation as a component.
     */
    public Component getContent() {
	return this;
    }

    public JProgressBar getProgress() {
	// Currently not available. Will have to do some magic using the
	// view.getStream() function.
	return null;
    }

    /** When a hyperlink event is fired we handle it in one of several ways,
     * based on the event type. Enters and exits get fired off as messages to
     * the listeners so they can update their status bars. Clicks also fire
     * status changes, but their event messages include the GURLs incase we
     * to do something special to them. Furthermore if follow hyperlinks is
     * set the browser is responsible for loading up the newly designated url.
     * @param event A HyperlinkEvent containing useful data about the event,
     * such as the url in question and the position of the mouse. It also is
     * marked as one of several types of events including ENTERED and EXITED
     * (for mouseovers) and ACTIVATED (clicked).
     */
    public void hyperlinkUpdate(HyperlinkEvent event) {
	Object event_type = event.getEventType();
	GURL new_url = new GURL(event.getURL());
	// The mouse has been moved over a hyperlink.
	if (event_type == HyperlinkEvent.EventType.ENTERED) {
	    fireStatusChange(new GURL(event.getURL()));
	} 
	else if (event_type == HyperlinkEvent.EventType.EXITED) {
	    fireStatusChange(null);
	}
	else if (event_type == HyperlinkEvent.EventType.ACTIVATED) {
	    fireHyperlinkClick(new_url);
	    if(follow_links) {
		show(new_url);
	    }
	}
    }

    public boolean isReady() {
	return ready;
    }

    public void propertyChange(PropertyChangeEvent evt) {
	if(evt.getPropertyName() == "page") {
	    ready = true;
	    fireRenderComplete(current_url);
	}
    }

    public void removeGBrowserListener(GBrowserListener listener) {
	listeners.remove(GBrowserListener.class, listener);
    }

    public void setEnabled(boolean value) {
	ready = value;
    }

    public boolean show(String filename) {
	return show(new GURL("file:"+filename));
    }

    /**
     *
     * Precondition: ready = true <BR>
     */
    public boolean show(GURL url) {
	if(current_url == null) {
	    fireRenderBegun(url);
	    current_url = url;
	    setPage(current_url);
	    return true;
	}
	else {
				// If we are not already showing this url
	    if(!current_url.toString().equals(url.toString())) {
		fireRenderBegun(url);
		// Save a copy of the current page
		back_pages.add(current_url);
		current_url = url;
		setPage(current_url);
		return true;
	    }
	}
	return false;
    }

    public boolean stop() {
	// Unfortunately a JEditorPane is not sophisticated enough to support
	// this. Pretty much once you send the page to be rendered only killing
	// the thread will stop it, and I'm not sure you can restart the thread
	// later.
	return false;
    }

    protected void fireActionFailed(GURL url) {
	GBrowserEvent event = new GBrowserEvent(this,0,url);
	Object[] concerned = listeners.getListenerList();
	for(int i = 0; i < concerned.length ; i++) {
	    if(concerned[i] == GBrowserListener.class) {
		((GBrowserListener)concerned[i+1]).actionFailed(event);
	    }
	}
    }

    protected void fireControlsChanged() {
	Object[] concerned = listeners.getListenerList();
	for(int i = 0; i < concerned.length ; i++) {
	    if(concerned[i] == GBrowserListener.class) {
		((GBrowserListener)concerned[i+1]).controlsChanged();
	    }
	}
    }

    protected void fireHyperlinkClick(GURL url) {
	GBrowserEvent event = new GBrowserEvent(this,0,url);
	Object[] concerned = listeners.getListenerList();
	for(int i = 0; i < concerned.length ; i++) {
	    if(concerned[i] == GBrowserListener.class) {
		((GBrowserListener)concerned[i+1]).hyperlinkClick(event);
	    }
	}
    }

    protected void fireRenderBegun(GURL url) {
	GBrowserEvent event = new GBrowserEvent(this,0,url);
	Object[] concerned = listeners.getListenerList();
	for(int i = 0; i < concerned.length ; i++) {
	    if(concerned[i] == GBrowserListener.class) {
		((GBrowserListener)concerned[i+1]).renderBegun(event);
	    }
	}
    }

    protected void fireRenderComplete(GURL url) {
	GBrowserEvent event = new GBrowserEvent(this,0,url);
	Object[] concerned = listeners.getListenerList();
	for(int i = 0; i < concerned.length ; i++) {
	    if(concerned[i] == GBrowserListener.class) {
		((GBrowserListener)concerned[i+1]).renderComplete(event);
	    }
	}
    }				 

    protected void fireStatusChange(GURL url) {
	GBrowserEvent event = new GBrowserEvent(this,0,url);
	Object[] concerned = listeners.getListenerList();
	for(int i = 0; i < concerned.length ; i++) {
	    if(concerned[i] == GBrowserListener.class) {
		((GBrowserListener)concerned[i+1]).statusChanged(event);
	    }
	}
    }

    private void setPage(GURL url) {
	try {
	    view.setPage(url.getURL());
	}
	catch (IOException e) {
	    Utility.showError(gatherer, "GHTMLPane.setPage()", e);
	}
	catch (Exception e) {
	    Utility.showError(gatherer, "Arrggg...", e);
	}
    }
}






